﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.IO;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using TSLab.Script;
using TSLab.Script.Handlers;
using TSLab.Script.Handlers.Options;
using TSLab.Script.Realtime;

namespace TradesLogerScript
{
    [HandlerCategory("S.Gilyazov")]
    [HelperName("Trades loger", Language = "en-US")]
    [HelperName("Trades loger", Language = "ru-RU")]
    [InputsCount(1)]
    [Input(0, TemplateTypes.SECURITY, Name = "SECURITYSource")]
    [OutputsCount(0)]
    [Description(
        "Записывает данные о сделках в файл. Один и тот же файл может использоваться для записи сделок нескольких агентов. " +
        "!!!ВАЖНО использование в сигналох входа/выхода, имени агента и т.д. символа | может привести к не верной записи данных. " +
        "Работает только в режиме агента, записывает данные о сделках закрытых на последнем баре. " +
        "Записывает следующие данные:" +
        "Имя агента; Tag1; Tag2; Направление; Инструмент; Максимальный размер позиции; Сигнал входа; Дата и время входа; Цена входа; Сигнал выхода; Дата и время выхода; Цена выхода; Прибыль(пункты); Прибыль(деньги); Прибыль %; Счет;" +
        "Прибыль в пунктах и процентах рассчитывается тслаб.Прибыль в деньгах рассчитывается как Прибыль в пунктах / шаг цены * стоимость шага цены" )] 
        
    
    public class TradesLoger : IStreamHandler, IOneSourceHandler, ISecurityInput0, IContextUses
    {
        public IContext Context { get; set; }

        /// <summary>
        /// \~english Path to log file
        /// \~russian Путь к логу 
        /// </summary>
        [HelperName("Path", "en-us")]
        [HelperName("Путь", "ru-ru")]
        [Description(@"Путь к файлу.Вместо \ необходимо писать \\")]
        [HelperDescription(@"Путь к файлу. Вместо \ необходимо писать \\", "en-us")]
        [HandlerParameter(true, "", NotOptimized = true)]
        public string FilePath { get; set; }


        /// <summary>
        /// \~russian Использовать торговое имя агента вместо вводимого в параметрах кубика
        /// </summary>
        [Description("Использовать торговое имя агента вместо вводимого в параметрах кубика")]
        [HandlerParameter(true, "", NotOptimized = true, Default = "false")]
        public bool UseAgentTradeName { get; set; }

        /// <summary>
        /// \~english Trade agent name
        /// \~russian Имя торгового агента
        /// </summary>
        [HelperName("Trade agent name", "en-us")]
        [HelperName("Имя торгового агента", "ru-ru")]
        [Description("Имя торгового агента. Используется если не включена опция UseAgentTradeName")]
        [HelperDescription("Trade agent name", "en-us")]
        [HandlerParameter(true, "", NotOptimized = true, Default = "-")]
        public string TradeName { get; set; }

        /// <summary>
        /// \~english Tag1
        /// \~russian Tag1
        /// </summary>
        [HelperName("Tag1", "en-us")]
        [HelperName("Tag1", "ru-ru")]
        [Description("Дополнительный столбец для групирования агентов для удобства фильтрации сделок. Например 'трендовые' и 'контртрендовые'")]
        [HelperDescription("Tag1", "en-us")]
        [HandlerParameter(true, "", NotOptimized = true, Default = "-")]
        public string Tag1 { get; set; }

        /// <summary>
        /// \~english Tag2
        /// \~russian Tag2
        /// </summary>
        [HelperName("Tag2", "en-us")]
        [HelperName("Tag2", "ru-ru")]
        [Description("Лень копипастиь. Это слишком сложно")]
        [HelperDescription("Tag2", "en-us")]
        [HandlerParameter(true, "", NotOptimized = true, Default = "-")]
        public string Tag2 { get; set; }
       

        /// <summary>
        /// \~english Message to TSLab log
        /// \~russian Выводить сообщение в лог ТСЛаб
        /// </summary>
        [Description("Выводить в лог ТСлаб сообщение аналогичное записываемому в файл для дальнейшей пересылки через менедджер оповещений")]
        [HandlerParameter(true, "", NotOptimized = true, Default = "false")]
        public bool MessageToLog { get; set; }


        /// <summary>
        /// \~english Message importance (Info, Warning, Error)
        /// \~russian Важность сообщения (Info, Warning, Error)
        /// </summary>
        [HelperName("Importance", "en-us")]
        [HelperName("Важность", "ru-ru")]
        [Description("Важность сообщения (Info, Warning, Error)")]
        [HelperDescription("Message importance (Info, Warning, Error)", "en-us")]
        [HandlerParameter(true, "Info", NotOptimized = true)]
        public MessageType Type { get; set; }

        /// <summary>
        /// \~english Additional user tag
        /// \~russian Дополнительная пользовательская метка
        /// </summary>
        [HelperName("MessageTag", "en-us")]
        [HelperName("Метка сообщения", "ru-ru")]
        [Description("Дополнительная пользовательская метка")]
        [HelperDescription("Additional user tag", "en-us")]
        [HandlerParameter(true, "MessageTag", NotOptimized = true)]
        public string MessageTag { get; set; }


        public static void TradeLogingWriter(List<string> messageList, IContext ctx, string filePath)
        {
            if (!File.Exists(filePath))
            {
                using (StreamWriter sw = new StreamWriter(filePath, true, Encoding.GetEncoding("windows-1251")))
                {
                    sw.WriteLine(
                        "Имя агента;Tag1;Tag2;Направление;Инструмент;Максимальный размер позиции;Сигнал входа;Дата и время входа;Цена входа;Сигнал выхода;Дата и время выхода;Цена выхода; Прибыль(пункты);Прибыль(деньги);Прибыль %;Счет;");
                }
            }

            try
            {
                using (StreamWriter sw = new StreamWriter(filePath, true, Encoding.GetEncoding("windows-1251")))
                {
                    foreach (var message in messageList)
                        sw.WriteLine(message);
                }
            }
            catch (Exception e)
            {
                ctx.Log($"Не удалось записать сделку в лог. Сохраняем ее в кэш для последующей записи. Exception {e}");
                var obj = ctx.LoadObject("TradesLoging postponed trades", true);
                StringBuilder allMessages = new StringBuilder();
                if (obj != null)
                    allMessages.Append(obj.ToString());

                foreach (var message in messageList)
                    allMessages.Append(message + '|');

                ctx.StoreObject("TradesLoging postponed trades", allMessages, true);
            }
        }

        public void Execute(ISecurity sec)
        {
            if (!Context.IsLastBarClosed)
                return;

            if (!Context.IsLastBarUsed)
                return;

            if (!Context.Runtime.IsAgentMode)
                return;
           
            Dictionary<string, object> dictionary = new Dictionary<string, object>()
            {
                {
                    "$UserMessageTag",
                    (object) (this.MessageTag ?? string.Empty)
                }
            };

            TradeName = TradeName == null ? "" : TradeName;
            Tag1 = Tag1 == null ? "" : Tag1;
            Tag2 = Tag2 == null ? "" : Tag2;
            string tradeName = UseAgentTradeName ? Context.Runtime.TradeName : TradeName;

            if (UseAgentTradeName)
            {
                
            }
            #region запись отложенных данных
            var obj = Context.LoadObject("TradesLoging postponed trades", true);
            Context.StoreObject("TradesLoging postponed trades", "", true);


            if (obj != null && obj.ToString() != "")
            {
                Context.Log($"Loger: записываю отложенные данные.", MessageType.Warning, true);
                var msgArray = obj.ToString().Split('|').ToList();
                msgArray.RemoveAt(msgArray.Count - 1);
                TradeLogingWriter(msgArray, Context, FilePath);
            }

            #endregion

            #region контроль повторной записи
            obj = Context.LoadObject("TradesLoging LastLogingTime", true);

            var time = DateTime.MinValue; // дата и время последней попытки записать прибыль
            if (obj != null)
                DateTime.TryParse(obj.ToString(), out time);
          
            if (time == sec.Bars[Context.BarsCount - 1].Date)
                return;
            #endregion

            var posList = sec.Positions.GetClosedForBar(Context.BarsCount - 1);
            var msgList = new List<string>();

            foreach (var pos in posList)
            {
                if (!pos.IsActiveForBar(Context.BarsCount - 1) && pos.IsActiveForBar(Context.BarsCount - 2) && !pos.IsVirtual && !pos.IsVirtualClosed)
                {
                    var direction = pos.IsLong ? "Long" : "Short";
                    var profit = pos.Profit();
                    var profitMoney = pos.Profit() / sec.Tick * sec.FinInfo.StepPrice;
                     
                    msgList.Add(tradeName + ";"
                                               + Tag1 + ";"
                                               + Tag2 + ";"
                                               + direction + ";"
                                               + sec.SecurityDescription.FullName + ";"
                                               + pos.MaxShares + ";"
                                               + pos.EntrySignalName + ";"
                                               + sec.Bars[pos.EntryBarNum].Date + ";"
                                               + pos.EntryPrice + ";"
                                               + pos.ExitSignalName + ";"
                                               + sec.Bars[pos.ExitBarNum].Date + ";"
                                               + pos.ExitPrice + ";"
                                               + profit + ";"
                                               + profitMoney + ";"
                                               + pos.ProfitPct() + ";"
                                               + (sec as ISecurityRt).PortfolioName + ";");
                    Context.StoreObject("TradesLoging LastLogingTime", sec.Bars[Context.BarsCount - 1].Date, true);
                    if (MessageToLog)
                        Context.Log("Loger: закрыта позиция. " + msgList.Last(),Type, true, (IDictionary<string, object>)dictionary);
                }
            }

            if (msgList.Any())
                TradeLogingWriter(msgList, Context, FilePath);
        }
    }
}